#  Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Convolutional Neural Network Estimator for MNIST, built with tf.layers."""

from __future__ import absolute_import, division, print_function

from nninst import mode
from nninst.backend.tensorflow.dataset import imagenet_raw
from nninst.backend.tensorflow.dataset.config import (
    CIFAR10_TRAIN_VGG16,
    IMAGENET_RAW_TRAIN,
)
from nninst.backend.tensorflow.model.config import VGG_16, VGG_16_CIFAR10
from nninst.statistics import calc_density_compact
from nninst.trace import TraceKey
from nninst.utils.ray import ray_init

from .common import (
    class_trace,
    class_trace_compact,
    full_trace,
    save_class_traces,
    save_class_traces_low_latency,
    self_similarity,
)

__all__ = ["vgg_16_cifar10_class_trace", "vgg_16_cifar10_self_similarity"]

name = "vgg_16_cifar10"

vgg_16_cifar10_class_trace = class_trace(
    name=name,
    model_config=VGG_16_CIFAR10,
    data_config=CIFAR10_TRAIN_VGG16,
    use_raw=True,
)

vgg_16_cifar10_class_trace_compact = class_trace_compact(
    vgg_16_cifar10_class_trace, name=name, model_config=VGG_16_CIFAR10
)

save_vgg_16_cifar10_class_traces_low_latency = save_class_traces_low_latency(
    name=name,
    model_config=VGG_16_CIFAR10,
    data_config=CIFAR10_TRAIN_VGG16,
    use_raw=True,
)

vgg_16_cifar10_trace = full_trace(name=name, class_trace_fn=vgg_16_cifar10_class_trace)

vgg_16_cifar10_self_similarity = self_similarity(
    name=name, trace_fn=vgg_16_cifar10_class_trace, class_ids=range(0, 10)
)

if __name__ == "__main__":
    # mode.check(False)
    mode.debug()
    # mode.local()
    # mode.distributed()
    # ray_init("dell")
    ray_init()
    # ray_init("r730")
    threshold = 0.5
    # threshold = 1
    # threshold = 0.8

    label = None

    print(f"generate class trace for label {label}")

    save_class_traces(
        vgg_16_cifar10_class_trace,
        range(10),
        threshold=threshold,
        label=label,
        example_num=10,
        example_upperbound=10,
        parallel=4,
        # cache=False,
    )

    # save_vgg_16_cifar10_class_traces_low_latency(
    #     np.transpose(np.reshape(np.array(range(0, 1000)), (100, 10))).flatten(),
    #     threshold=threshold, label=label,
    #     example_num=100, example_upperbound=1000,
    #     merge_fn=merge_simple_trace)

    # check_class_traces(vgg_16_cifar10_class_trace, range(0, 1000), threshold, label, compress=True)

    # save_class_traces(vgg_16_cifar10_class_trace_compact, range(0, 1000), threshold=threshold, label=label)

    # vgg_16_cifar10_self_similarity(threshold, label).save()

    # trace = vgg_16_cifar10_trace(threshold=threshold, label=label).load()
    # layers = VGG16.graph().load().layers()
    # for key in [TraceKey.POINT, TraceKey.EDGE, TraceKey.WEIGHT]:
    #     density_per_layer = calc_density_compact_per_layer(trace, layers, key)
    #     density_per_layer.to_csv(abspath(f"vgg_16_cifar10_trace_per_layer.{key}.csv"))

    # trace = vgg_16_cifar10_trace(threshold, label).load()
    # for key in [TraceKey.POINT, TraceKey.EDGE, TraceKey.WEIGHT]:
    #     print(f"{key}: {calc_density_compact(trace, key)}")
